/* mapview - map viewer for PalmOS
   Copyright (C) 2005  Edwin Steiner <edwin.steiner@gmx.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <PalmOS.h>
#include "view.h"
#include "memory.h"
#include "debug.h"

mapview_view_t *
mapview_view_new(mapview_map_t *map,UInt16 coordsys,int vx,int vy,int sx,int sy)
{
        mapview_view_t *view;

        ASSERT(map);
        ASSERT(sx >= 0);
        ASSERT(sy >= 0);

        view = NEW(mapview_view_t);
        if (!view)
                return NULL;

        view->map       = map;
        view->x         = 0; /* XXX center? */
        view->y         = 0;
        view->layer     = 0;
        view->coordsys  = coordsys;
        view->viewx     = vx;
        view->viewy     = vy;
        view->viewsizex = sx;
        view->viewsizey = sy;
        view->stepsizex = sx / 4;
        view->stepsizey = sy / 4;

        view->xmin      = 0;
        view->ymin      = 0;
        view->xmax      = 0;
        view->ymax      = 0;

        if (view->stepsizex < 1)
                view->stepsizex = 1;
        if (view->stepsizey < 1)
                view->stepsizey = 1;

        mapview_view_switch_to_layer(view,0);

        return view;
}

void
mapview_view_free(mapview_view_t *view)
{
        if (!view)
                return;

        FREE(mapview_view_t,view);
}

void 
mapview_view_draw(mapview_view_t *view)
{
        Int32 x,y,ex,ey,sx,sy,tx,ty;
        BitmapPtr bmp;
        RectangleType rect;
        UInt16 coordsys;
        mapview_layer_t *layer;
        RGBColorType rgb;
        
        ASSERT(view);

        if (g_high_density_feature) {
                coordsys = WinSetCoordinateSystem(view->coordsys);
        }

        if (g_rom_version >= 0x03503000) {
                rgb.index = 0;
                rgb.r = rgb.g = rgb.b = 0;
                WinSetForeColor(WinRGBToIndex(&rgb));
        }

        layer = view->map->layers[view->layer];
        x = view->x / layer->scale;
        y = view->y / layer->scale;
        ex = x + view->viewsizex;
        ey = y + view->viewsizey;

        if (y >= 0) {
                sy = (y / layer->tilesizey) * layer->tilesizey;
        }
        else {
                sy = ((y-layer->tilesizey+1) / layer->tilesizey) * layer->tilesizey;
        }
        while (sy < ey) {
                ty = sy / layer->tilesizey;
                if (x >= 0) {
                        sx = (x / layer->tilesizex) * layer->tilesizex;
                }
                else {
                        sx = ((x-layer->tilesizex+1) / layer->tilesizex) * layer->tilesizex;
                }
                while (sx < ex) {
                        tx = sx / layer->tilesizex;
                        bmp = mapview_layer_get_tile(layer,tx,ty);
                        if (bmp) {
                                WinDrawBitmap(bmp,view->viewx + sx-x,view->viewy + sy-y);
                                mapview_layer_release_tile(layer);
                        }
                        else {
                                rect.topLeft.x = view->viewx + sx - x;
                                rect.topLeft.y = view->viewy + sy - y;
                                rect.extent.x = layer->tilesizex;
                                rect.extent.y = layer->tilesizey;
                                WinDrawRectangle(&rect,0);
                        }
                        sx += layer->tilesizex;
                }
                sy += layer->tilesizey;
        }
        if (g_high_density_feature) {
                WinSetCoordinateSystem(coordsys);
        }
}

Int32 
mapview_view_scale(mapview_view_t *view)
{
        ASSERT(view);

        return view->map->layers[view->layer]->scale;
}

void 
mapview_view_get_center(mapview_view_t *view,Int32 *px,Int32 *py)
{
        Int32 scale;
        
        ASSERT(view);
        ASSERT(px);
        ASSERT(py);

        scale = mapview_view_scale(view);
        *px = view->x + scale * view->viewsizex / 2;
        *py = view->y + scale * view->viewsizey / 2;
}

void 
mapview_view_get_visible_rect(mapview_view_t *view,Int32 *x,Int32 *y,Int32 *ex,Int32 *ey)
{
        Int32 scale;
        
        ASSERT(view);
        ASSERT(x);
        ASSERT(y);
        ASSERT(ex);
        ASSERT(ey);

        scale = mapview_view_scale(view);
        *x = view->x;
        *y = view->y;
        *ex = view->x + scale * view->viewsizex;
        *ey = view->y + scale * view->viewsizey;
}

static void
mapview_view_std_to_native(mapview_view_t *view,Int32 *x,Int32 *y)
{
        UInt16 coordsys;
        PointType point;

        ASSERT(view);
        ASSERT(x);
        ASSERT(y);

        if (g_high_density_feature) {
                coordsys = WinSetCoordinateSystem(view->coordsys);
                point.x = *x;
                point.y = *y;
                WinScalePoint(&point,false);
                *x = point.x;
                *y = point.y;
                WinSetCoordinateSystem(coordsys);
        }
}

void 
mapview_view_screen_to_map(mapview_view_t *view,Int32 screenx,Int32 screeny,
                                                Int32 *x,Int32 *y)
{
        Int32 scale;

        ASSERT(view);
        ASSERT(x);
        ASSERT(y);

        scale = mapview_view_scale(view);

        /* convert to coordsys */
        mapview_view_std_to_native(view,&screenx,&screeny);

        /* XXX coords relative to top-left corner of view */
        screenx -= view->viewx;
        screeny -= view->viewy;

        *x = view->x + scale*screenx;
        *y = view->y + scale*screeny;
}

bool 
mapview_view_move_to(mapview_view_t *view,Int32 x,Int32 y)
{
        Int32 scale;
        
        ASSERT(view);
        
        scale = mapview_view_scale(view);
        if (x < view->xmin)
                x = view->xmin;
        if (y < view->ymin)
                y = view->ymin;
        if (x > view->xmax)
                x = view->xmax;
        if (y > view->ymax)
                y = view->ymax;

        if (x != view->x || y != view->y) {
                view->x = x;
                view->y = y;
                return true;
        }
        return false;
}

bool 
mapview_view_move_to_map_center(mapview_view_t *view)
{
        ASSERT(view);
        return mapview_view_move_center(view,
                        view->map->mapsizex / 2,
                        view->map->mapsizey / 2);
}

bool 
mapview_view_move_center(mapview_view_t *view,Int32 x,Int32 y)
{
        Int32 scale;
        
        ASSERT(view);
        
        scale = mapview_view_scale(view);
        return mapview_view_move_to(view,
                        x - scale*view->viewsizex/2,
                        y - scale*view->viewsizey/2);
}

bool 
mapview_view_move(mapview_view_t *view,Int32 dx,Int32 dy)
{
        return mapview_view_move_to(view,view->x + dx,view->y + dy);
}

bool 
mapview_view_move_scaled(mapview_view_t *view,Int32 dx,Int32 dy)
{
        Int32 scale;
        
        ASSERT(view);
        
        scale = mapview_view_scale(view);
        return mapview_view_move(view, dx * scale, dy * scale);
}

bool 
mapview_view_down(mapview_view_t *view)
{
        return mapview_view_switch_to_layer(view,view->layer + 1);
}

bool 
mapview_view_up(mapview_view_t *view)
{
        return mapview_view_switch_to_layer(view,view->layer - 1);
}

bool 
mapview_view_switch_to_layer(mapview_view_t *view,int layer)
{
        Int32 x,y;
        Int32 scale;
        
        ASSERT(view);
        
        if (layer < 0 || layer >= view->map->nlayers)
                return false;

        mapview_view_get_center(view,&x,&y);
        view->layer = layer;

        scale = mapview_view_scale(view);
        view->xmin = - scale*view->viewsizex/2;
        view->ymin = - scale*view->viewsizey/2;
        view->xmax = view->map->mapsizex - scale*view->viewsizex/2;
        view->ymax = view->map->mapsizey - scale*view->viewsizey/2;

        mapview_view_move_center(view,x,y);

        return true;
}

/* vim: et sts=8 sw=8 ai
 */
